from typing import Any, Optional
import pytest
import eagerpy as ep


def pytest_addoption(parser: Any) -> None:
    parser.addoption("--backend")


@pytest.fixture(scope="session")
def dummy(request: Any) -> ep.Tensor:
    backend: Optional[str] = request.config.option.backend
    if backend is None:
        pytest.skip()
        assert False
    return ep.utils.get_dummy(backend)


@pytest.fixture(scope="session")
def t1(dummy: ep.Tensor) -> ep.Tensor:
    return ep.arange(dummy, 5).float32()


@pytest.fixture(scope="session")
def t1int(dummy: ep.Tensor) -> ep.Tensor:
    return ep.arange(dummy, 5)


@pytest.fixture(scope="session")
def t2(dummy: ep.Tensor) -> ep.Tensor:
    return ep.arange(dummy, 7, 17, 2).float32()


@pytest.fixture(scope="session")
def t2int(dummy: ep.Tensor) -> ep.Tensor:
    return ep.arange(dummy, 7, 17, 2)


@pytest.fixture(scope="session", params=["t1", "t2"])
def t(request: Any, t1: ep.Tensor, t2: ep.Tensor) -> ep.Tensor:
    return {"t1": t1, "t2": t2}[request.param]



from typing import overload, Sequence, Callable, Tuple, Any, Optional, cast, Union
from typing_extensions import Literal

from .types import Axes, AxisAxes, Shape, ShapeOrScalar

from .tensor import Tensor
from .tensor import TensorType
from .tensor import TensorOrScalar

newaxis = None
inf = float("inf")
nan = float("nan")


def clip(t: TensorType, min_: float, max_: float) -> TensorType:
    return t.clip(min_, max_)


def abs(t: TensorType) -> TensorType:
    return t.abs()


def sign(t: TensorType) -> TensorType:
    return t.sign()


def sqrt(t: TensorType) -> TensorType:
    return t.sqrt()


def square(t: TensorType) -> TensorType:
    return t.square()


def pow(t: TensorType, exponent: TensorOrScalar) -> TensorType:
    return t.pow(exponent)


def tanh(t: TensorType) -> TensorType:
    return t.tanh()


def arctanh(t: TensorType) -> TensorType:
    return t.arctanh()


def sum(
    t: TensorType, axis: Optional[AxisAxes] = None, keepdims: bool = False
) -> TensorType:
    return t.sum(axis=axis, keepdims=keepdims)


def prod(
    t: TensorType, axis: Optional[AxisAxes] = None, keepdims: bool = False
) -> TensorType:
    return t.prod(axis=axis, keepdims=keepdims)


def mean(
    t: TensorType, axis: Optional[AxisAxes] = None, keepdims: bool = False
) -> TensorType:
    return t.mean(axis=axis, keepdims=keepdims)


def min(
    t: TensorType, axis: Optional[AxisAxes] = None, keepdims: bool = False
) -> TensorType:
    return t.min(axis=axis, keepdims=keepdims)


def max(
    t: TensorType, axis: Optional[AxisAxes] = None, keepdims: bool = False
) -> TensorType:
    return t.max(axis=axis, keepdims=keepdims)


@overload
def minimum(x: TensorType, y: TensorOrScalar) -> TensorType:
    ...


@overload
def minimum(x: TensorOrScalar, y: TensorType) -> TensorType:
    ...


def minimum(x: TensorOrScalar, y: TensorOrScalar) -> Tensor:
    if not isinstance(x, Tensor):
        return cast(Tensor, y).minimum(x)
    return x.minimum(y)


@overload
def maximum(x: TensorType, y: TensorOrScalar) -> TensorType:
    ...


@overload
def maximum(x: TensorOrScalar, y: TensorType) -> TensorType:
    ...


def maximum(x: TensorOrScalar, y: TensorOrScalar) -> Tensor:
    if not isinstance(x, Tensor):
        return cast(Tensor, y).maximum(x)
    return x.maximum(y)


def argmin(t: TensorType, axis: Optional[int] = None) -> TensorType:
    return t.argmin(axis=axis)


def argmax(t: TensorType, axis: Optional[int] = None) -> TensorType:
    return t.argmax(axis=axis)


def argsort(t: TensorType, axis: int = -1) -> TensorType:
    return t.argsort(axis=axis)


def sort(t: TensorType, axis: int = -1) -> TensorType:
    return t.sort(axis=axis)


def topk(t: TensorType, k: int, sorted: bool = True) -> Tuple[TensorType, TensorType]:
    return t.topk(k, sorted=sorted)


def uniform(
    t: TensorType, shape: ShapeOrScalar, low: float = 0.0, high: float = 1.0
) -> TensorType:
    return t.uniform(shape, low=low, high=high)


def normal(
    t: TensorType, shape: ShapeOrScalar, mean: float = 0.0, stddev: float = 1.0
) -> TensorType:
    return t.normal(shape, mean=mean, stddev=stddev)


def ones(t: TensorType, shape: ShapeOrScalar) -> TensorType:
    return t.ones(shape)


def zeros(t: TensorType, shape: ShapeOrScalar) -> TensorType:
    return t.zeros(shape)


def ones_like(t: TensorType) -> TensorType:
    return t.ones_like()


def zeros_like(t: TensorType) -> TensorType:
    return t.zeros_like()


def full_like(t: TensorType, fill_value: float) -> TensorType:
    return t.full_like(fill_value)


def onehot_like(t: TensorType, indices: TensorType, *, value: float = 1) -> TensorType:
    return t.onehot_like(indices, value=value)


def from_numpy(t: TensorType, a: Any) -> TensorType:
    return t.from_numpy(a)


def concatenate(tensors: Sequence[TensorType], axis: int = 0) -> TensorType:
    t = tensors[0]
    return t._concatenate(tensors, axis=axis)


def transpose(t: TensorType, axes: Optional[Axes] = None) -> TensorType:
    return t.transpose(axes=axes)


@overload
def logical_and(x: TensorType, y: TensorOrScalar) -> TensorType:
    ...


@overload
def logical_and(x: TensorOrScalar, y: TensorType) -> TensorType:
    ...


def logical_and(x: TensorOrScalar, y: TensorOrScalar) -> Tensor:
    if not isinstance(x, Tensor):
        return cast(Tensor, y).logical_and(x)
    return x.logical_and(y)


@overload
def logical_or(x: TensorType, y: TensorOrScalar) -> TensorType:
    ...


@overload
def logical_or(x: TensorOrScalar, y: TensorType) -> TensorType:
    ...


def logical_or(x: TensorOrScalar, y: TensorOrScalar) -> Tensor:
    if not isinstance(x, Tensor):
        return cast(Tensor, y).logical_or(x)
    return x.logical_or(y)


def logical_not(t: TensorType) -> TensorType:
    return t.logical_not()


def exp(t: TensorType) -> TensorType:
    return t.exp()


def log(t: TensorType) -> TensorType:
    return t.log()


def log2(t: TensorType) -> TensorType:
    return t.log2()


def log10(t: TensorType) -> TensorType:
    return t.log10()


def log1p(t: TensorType) -> TensorType:
    return t.log1p()


def where(condition: TensorType, x: TensorOrScalar, y: TensorOrScalar) -> TensorType:
    return condition.where(x, y)


def tile(t: TensorType, multiples: Axes) -> TensorType:
    return t.tile(multiples)


def matmul(x: TensorType, y: TensorType) -> TensorType:
    return x.matmul(y)


def softmax(t: TensorType, axis: int = -1) -> TensorType:
    return t.softmax(axis=axis)


def log_softmax(t: TensorType, axis: int = -1) -> TensorType:
    return t.log_softmax(axis=axis)


def stack(tensors: Sequence[TensorType], axis: int = 0) -> TensorType:
    t = tensors[0]
    return t._stack(tensors, axis=axis)


def squeeze(t: TensorType, axis: Optional[AxisAxes] = None) -> TensorType:
    return t.squeeze(axis=axis)


def expand_dims(t: TensorType, axis: int) -> TensorType:
    return t.expand_dims(axis=axis)


def full(t: TensorType, shape: ShapeOrScalar, value: float) -> TensorType:
    return t.full(shape, value)


def index_update(t: TensorType, indices: Any, values: TensorOrScalar) -> TensorType:
    return t.index_update(indices, values)


def arange(
    t: TensorType, start: int, stop: Optional[int] = None, step: Optional[int] = None
) -> TensorType:
    return t.arange(start, stop, step)


def cumsum(t: TensorType, axis: Optional[int] = None) -> TensorType:
    return t.cumsum(axis=axis)


def flip(t: TensorType, axis: Optional[AxisAxes] = None) -> TensorType:
    return t.flip(axis=axis)


def meshgrid(
    t: TensorType, *tensors: TensorType, indexing: str = "xy"
) -> Tuple[TensorType, ...]:
    return t.meshgrid(*tensors, indexing=indexing)


def pad(
    t: TensorType,
    paddings: Tuple[Tuple[int, int], ...],
    mode: str = "constant",
    value: float = 0,
) -> TensorType:
    return t.pad(paddings, mode=mode, value=value)


def isnan(t: TensorType) -> TensorType:
    return t.isnan()


def isinf(t: TensorType) -> TensorType:
    return t.isinf()


def all(
    t: TensorType, axis: Optional[AxisAxes] = None, keepdims: bool = False
) -> TensorType:
    return t.all(axis=axis, keepdims=keepdims)


def any(
    t: TensorType, axis: Optional[AxisAxes] = None, keepdims: bool = False
) -> TensorType:
    return t.any(axis=axis, keepdims=keepdims)


def crossentropy(logits: TensorType, labels: TensorType) -> TensorType:
    return logits.crossentropy(labels)


def slogdet(matrix: TensorType) -> Tuple[TensorType, TensorType]:
    return matrix.slogdet()


@overload
def value_and_grad_fn(
    t: TensorType, f: Callable[..., TensorType]
) -> Callable[..., Tuple[TensorType, TensorType]]:
    ...


@overload
def value_and_grad_fn(
    t: TensorType, f: Callable[..., TensorType], has_aux: Literal[False]
) -> Callable[..., Tuple[TensorType, TensorType]]:
    ...


@overload
def value_and_grad_fn(
    t: TensorType, f: Callable[..., Tuple[TensorType, Any]], has_aux: Literal[True]
) -> Callable[..., Tuple[TensorType, Any, TensorType]]:
    ...


def value_and_grad_fn(t: Any, f: Any, has_aux: bool = False) -> Any:
    return t._value_and_grad_fn(f, has_aux=has_aux)


def value_and_grad(
    f: Callable[..., TensorType], t: TensorType, *args: Any, **kwargs: Any
) -> Tuple[TensorType, TensorType]:
    return t.value_and_grad(f, *args, **kwargs)


def value_aux_and_grad(
    f: Callable[..., Tuple[TensorType, Any]], t: TensorType, *args: Any, **kwargs: Any
) -> Tuple[TensorType, Any, TensorType]:
    return t.value_aux_and_grad(f, *args, **kwargs)


def reshape(t: TensorType, shape: Union[Shape, int]) -> TensorType:
    return t.reshape(shape)


def take_along_axis(t: TensorType, indices: TensorType, axis: int) -> TensorType:
    return t.take_along_axis(indices, axis)


def flatten(t: TensorType, start: int = 0, end: int = -1) -> TensorType:
    return t.flatten(start=start, end=end)



from typing import Optional
import pytest
from numpy.testing import assert_allclose
from numpy.linalg import norm
import numpy as np
import eagerpy as ep
from eagerpy import Tensor
from eagerpy.norms import l0, l1, l2, linf, lp

norms = {0: l0, 1: l1, 2: l2, ep.inf: linf}


@pytest.fixture
def x1d(dummy: Tensor) -> Tensor:
    return ep.arange(dummy, 10).float32() / 7.0


@pytest.fixture
def x2d(dummy: Tensor) -> Tensor:
    return ep.arange(dummy, 12).float32().reshape((3, 4)) / 7.0


@pytest.fixture
def x4d(dummy: Tensor) -> Tensor:
    return ep.arange(dummy, 2 * 3 * 4 * 5).float32().reshape((2, 3, 4, 5)) / 7.0


@pytest.mark.parametrize("p", [0, 1, 2, ep.inf])
def test_1d(x1d: Tensor, p: float) -> None:
    assert_allclose(lp(x1d, p).numpy(), norm(x1d.numpy(), ord=p))
    assert_allclose(norms[p](x1d).numpy(), norm(x1d.numpy(), ord=p))


@pytest.mark.parametrize("p", [0, 1, 2, 3, 4, ep.inf])
@pytest.mark.parametrize("axis", [0, 1, -1])
@pytest.mark.parametrize("keepdims", [False, True])
def test_2d(x2d: Tensor, p: float, axis: int, keepdims: bool) -> None:
    assert isinstance(axis, int)  # see test4d for the more general test
    assert_allclose(
        lp(x2d, p, axis=axis, keepdims=keepdims).numpy(),
        norm(x2d.numpy(), ord=p, axis=axis, keepdims=keepdims),
        rtol=1e-6,
    )
    if p not in norms:
        return
    assert_allclose(
        norms[p](x2d, axis=axis, keepdims=keepdims).numpy(),
        norm(x2d.numpy(), ord=p, axis=axis, keepdims=keepdims),
        rtol=1e-6,
    )


@pytest.mark.parametrize("p", [0, 1, 2, 3, 4, ep.inf])
@pytest.mark.parametrize(
    "axis",
    [
        None,
        0,
        1,
        2,
        3,
        -1,
        -2,
        -3,
        -4,
        (0, 1),
        (1, 2),
        (1, 3),
        (1, 2, 3),
        (0, 1, 3),
        (2, 1, 0),
    ],
)
@pytest.mark.parametrize("keepdims", [False, True])
def test_4d(
    x4d: Tensor, p: float, axis: Optional[ep.types.AxisAxes], keepdims: bool
) -> None:
    actual = lp(x4d, p, axis=axis, keepdims=keepdims).numpy()

    # numpy does not support arbitrary axes (limited to vector and matrix norms)
    if axis is None:
        axes = tuple(range(x4d.ndim))
    elif not isinstance(axis, tuple):
        axes = (axis,)
    else:
        axes = axis
    del axis
    axes = tuple(i % x4d.ndim for i in axes)
    x = x4d.numpy()
    other = tuple(i for i in range(x.ndim) if i not in axes)
    x = np.transpose(x, other + axes)
    x = x.reshape(x.shape[: len(other)] + (-1,))
    desired = norm(x, ord=p, axis=-1)
    if keepdims:
        shape = tuple(1 if i in axes else x4d.shape[i] for i in range(x4d.ndim))
        desired = desired.reshape(shape)

    assert_allclose(actual, desired, rtol=1e-6)



from typing import overload
from typing_extensions import Literal

from .tensor import Tensor
from .tensor import PyTorchTensor
from .tensor import TensorFlowTensor
from .tensor import JAXTensor
from .tensor import NumPyTensor

from . import modules


@overload
def get_dummy(framework: Literal["pytorch"]) -> PyTorchTensor:
    ...


@overload
def get_dummy(framework: Literal["tensorflow"]) -> TensorFlowTensor:
    ...


@overload
def get_dummy(framework: Literal["jax"]) -> JAXTensor:
    ...


@overload
def get_dummy(framework: Literal["numpy"]) -> NumPyTensor:
    ...


@overload
def get_dummy(framework: str) -> Tensor:
    ...


def get_dummy(framework: str) -> Tensor:
    x: Tensor
    if framework == "pytorch":
        x = modules.torch.zeros(0)
        assert isinstance(x, PyTorchTensor)
    elif framework == "pytorch-gpu":
        x = modules.torch.zeros(0, device="cuda:0")  # pragma: no cover
        assert isinstance(x, PyTorchTensor)  # pragma: no cover
    elif framework == "tensorflow":
        x = modules.tensorflow.zeros(0)
        assert isinstance(x, TensorFlowTensor)
    elif framework == "jax":
        x = modules.jax.numpy.zeros(0)
        assert isinstance(x, JAXTensor)
    elif framework == "numpy":
        x = modules.numpy.zeros(0)
        assert isinstance(x, NumPyTensor)
    else:
        raise ValueError(f"unknown framework: {framework}")  # pragma: no cover
    return x.float32()



from typing import Union, Optional

from .tensor import TensorType
from .types import AxisAxes
from .framework import inf


def l0(
    x: TensorType, axis: Optional[AxisAxes] = None, keepdims: bool = False
) -> TensorType:
    return (x != 0).sum(axis=axis, keepdims=keepdims).astype(x.dtype)


def l1(
    x: TensorType, axis: Optional[AxisAxes] = None, keepdims: bool = False
) -> TensorType:
    return x.abs().sum(axis=axis, keepdims=keepdims)


def l2(
    x: TensorType, axis: Optional[AxisAxes] = None, keepdims: bool = False
) -> TensorType:
    return x.square().sum(axis=axis, keepdims=keepdims).sqrt()


def linf(
    x: TensorType, axis: Optional[AxisAxes] = None, keepdims: bool = False
) -> TensorType:
    return x.abs().max(axis=axis, keepdims=keepdims)


def lp(
    x: TensorType,
    p: Union[int, float],
    axis: Optional[AxisAxes] = None,
    keepdims: bool = False,
) -> TensorType:
    if p == 0:
        return l0(x, axis=axis, keepdims=keepdims)
    if p == 1:
        return l1(x, axis=axis, keepdims=keepdims)
    if p == 2:
        return l2(x, axis=axis, keepdims=keepdims)
    if p == inf:
        return linf(x, axis=axis, keepdims=keepdims)
    return x.abs().pow(p).sum(axis=axis, keepdims=keepdims).pow(1.0 / p)



from .tensor import Tensor  # noqa: F401
from .tensor import TensorType  # noqa: F401
from .tensor import TensorOrScalar  # noqa: F401
from .tensor import istensor  # noqa: F401

from .pytorch import PyTorchTensor  # noqa: F401
from .tensorflow import TensorFlowTensor  # noqa: F401
from .numpy import NumPyTensor  # noqa: F401
from .jax import JAXTensor  # noqa: F401



from typing import TypeVar, Callable, Any, Generic
import typing
import functools

from .. import norms

from .tensor import Tensor


T = TypeVar("T")


def extensionmethod(f: Callable[..., T]) -> Callable[..., T]:
    @functools.wraps(f)
    def wrapper(self: Any, *args: Any, **kwargs: Any) -> Any:
        return f(self._instance, *args, **kwargs)

    return wrapper


class ExtensionMeta(type):
    def __new__(cls, name, bases, attrs):  # type: ignore
        if bases != ():
            # creating a subclass of ExtensionMethods
            # wrap the attributes with extensionmethod
            attrs = {
                k: extensionmethod(v) if not k.startswith("__") else v
                for k, v in attrs.items()
            }
        return super().__new__(cls, name, bases, attrs)


if hasattr(typing, "GenericMeta"):  # Python 3.6
    # workaround for https://github.com/python/typing/issues/449
    class GenericExtensionMeta(typing.GenericMeta, ExtensionMeta):
        pass


else:  # pragma: no cover  # Python 3.7 and newer

    class GenericExtensionMeta(ExtensionMeta):  # type: ignore
        pass


class ExtensionMethods(metaclass=GenericExtensionMeta):
    def __init__(self, instance: Tensor):
        self._instance = instance


T_co = TypeVar("T_co", bound=Tensor, covariant=True)


class NormsMethods(Generic[T_co], ExtensionMethods):
    l0: Callable[..., T_co] = norms.l0
    l1: Callable[..., T_co] = norms.l1
    l2: Callable[..., T_co] = norms.l2
    linf: Callable[..., T_co] = norms.linf
    lp: Callable[..., T_co] = norms.lp



from typing_extensions import final
from typing import Any, cast

from .tensor import Tensor
from .tensor import TensorType
from .tensor import TensorOrScalar


def unwrap_(*args: Any) -> Any:
    return tuple(t.raw if isinstance(t, Tensor) else t for t in args)


def unwrap1(t: Any) -> Any:
    return t.raw if isinstance(t, Tensor) else t


class BaseTensor(Tensor):
    __slots__ = "_raw"

    def __init__(self: TensorType, raw: Any):
        assert not isinstance(raw, Tensor)
        self._raw = raw

    @property
    def raw(self) -> Any:
        return self._raw

    @final
    def __repr__(self: TensorType) -> str:
        lines = repr(self.raw).split("\n")
        prefix = self.__class__.__name__ + "("
        lines[0] = prefix + lines[0]
        prefix = " " * len(prefix)
        for i in range(1, len(lines)):
            lines[i] = prefix + lines[i]
        lines[-1] = lines[-1] + ")"
        return "\n".join(lines)

    @final
    def __format__(self: TensorType, format_spec: str) -> str:
        return format(self.raw, format_spec)

    @final
    @property
    def dtype(self: TensorType) -> Any:
        return self.raw.dtype

    @final
    def __bool__(self: TensorType) -> bool:
        return bool(self.raw)

    @final
    def __len__(self: TensorType) -> int:
        return cast(int, self.raw.shape[0])

    @final
    def __abs__(self: TensorType) -> TensorType:
        return type(self)(abs(self.raw))

    @final
    def __neg__(self: TensorType) -> TensorType:
        return type(self)(-self.raw)

    @final
    def __add__(self: TensorType, other: TensorOrScalar) -> TensorType:
        return type(self)(self.raw.__add__(unwrap1(other)))

    @final
    def __radd__(self: TensorType, other: TensorOrScalar) -> TensorType:
        return type(self)(self.raw.__radd__(unwrap1(other)))

    @final
    def __sub__(self: TensorType, other: TensorOrScalar) -> TensorType:
        return type(self)(self.raw.__sub__(unwrap1(other)))

    @final
    def __rsub__(self: TensorType, other: TensorOrScalar) -> TensorType:
        return type(self)(self.raw.__rsub__(unwrap1(other)))

    @final
    def __mul__(self: TensorType, other: TensorOrScalar) -> TensorType:
        return type(self)(self.raw.__mul__(unwrap1(other)))

    @final
    def __rmul__(self: TensorType, other: TensorOrScalar) -> TensorType:
        return type(self)(self.raw.__rmul__(unwrap1(other)))

    @final
    def __truediv__(self: TensorType, other: TensorOrScalar) -> TensorType:
        return type(self)(self.raw.__truediv__(unwrap1(other)))

    @final
    def __rtruediv__(self: TensorType, other: TensorOrScalar) -> TensorType:
        return type(self)(self.raw.__rtruediv__(unwrap1(other)))

    @final
    def __floordiv__(self: TensorType, other: TensorOrScalar) -> TensorType:
        return type(self)(self.raw.__floordiv__(unwrap1(other)))

    @final
    def __rfloordiv__(self: TensorType, other: TensorOrScalar) -> TensorType:
        return type(self)(self.raw.__rfloordiv__(unwrap1(other)))

    @final
    def __mod__(self: TensorType, other: TensorOrScalar) -> TensorType:
        return type(self)(self.raw.__mod__(unwrap1(other)))

    @final
    def __pow__(self: TensorType, exponent: TensorOrScalar) -> TensorType:
        return type(self)(self.raw.__pow__(unwrap1(exponent)))

    @final
    @property
    def ndim(self: TensorType) -> int:
        return len(self.raw.shape)



from typing import TypeVar
from os.path import join as _join
from os.path import dirname as _dirname

with open(_join(_dirname(__file__), "VERSION")) as _f:
    __version__ = _f.read().strip()

_T = TypeVar("_T")


class _Indexable:
    __slots__ = ()

    def __getitem__(self, index: _T) -> _T:
        return index


index = _Indexable()


from .tensor import Tensor  # noqa: F401,E402
from .tensor import TensorType  # noqa: F401,E402
from .tensor import istensor  # noqa: F401,E402

from .tensor import PyTorchTensor  # noqa: F401,E402
from .tensor import TensorFlowTensor  # noqa: F401,E402
from .tensor import NumPyTensor  # noqa: F401,E402
from .tensor import JAXTensor  # noqa: F401,E402

from . import types  # noqa: F401,E402

from .astensor import astensor  # noqa: F401,E402
from .astensor import astensors  # noqa: F401,E402
from .astensor import astensor_  # noqa: F401,E402
from .astensor import astensors_  # noqa: F401,E402

from .modules import torch  # noqa: F401,E402
from .modules import tensorflow  # noqa: F401,E402
from .modules import jax  # noqa: F401,E402
from .modules import numpy  # noqa: F401,E402

from . import utils  # noqa: F401,E402

from .framework import *  # noqa: F401,E402,F403

from . import norms  # noqa: F401,E402
from .lib import *  # noqa: F401,E402,F403
